@@ -12,13 +12,17 @@ from fontTools.ttLib import TTFont, TTLibError
 from fontTools.varLib import instancer
 from fontTools.varLib.instancer import OverlapMode
 
+from PIL import Image, ImageDraw, ImageFont
+
 import copy
 import fsutil
+import imagehash
 import itertools
 import math
 import os
 import re
 import sys
+import tempfile
 
 
 class Font(object):
@@ -161,20 +165,21 @@ class Font(object):
     WEIGHT_EXTRA_BOLD = "Extra-bold"  # (Ultra-bold)
     WEIGHT_BLACK = "Black"  # (Heavy)
     WEIGHT_EXTRA_BLACK = "Extra-black"  # (Nord)
-    _WEIGHTS_BY_VALUE = {
-        50: WEIGHT_EXTRA_THIN,
-        100: WEIGHT_THIN,
-        200: WEIGHT_EXTRA_LIGHT,
-        300: WEIGHT_LIGHT,
-        400: WEIGHT_REGULAR,
-        450: WEIGHT_BOOK,
-        500: WEIGHT_MEDIUM,
-        600: WEIGHT_SEMI_BOLD,
-        700: WEIGHT_BOLD,
-        800: WEIGHT_EXTRA_BOLD,
-        900: WEIGHT_BLACK,
-        950: WEIGHT_EXTRA_BLACK,
-    }
+    _WEIGHTS = [
+        {"value": 50, "name": WEIGHT_EXTRA_THIN},
+        {"value": 100, "name": WEIGHT_THIN},
+        {"value": 200, "name": WEIGHT_EXTRA_LIGHT},
+        {"value": 300, "name": WEIGHT_LIGHT},
+        {"value": 400, "name": WEIGHT_REGULAR},
+        {"value": 450, "name": WEIGHT_BOOK},
+        {"value": 500, "name": WEIGHT_MEDIUM},
+        {"value": 600, "name": WEIGHT_SEMI_BOLD},
+        {"value": 700, "name": WEIGHT_BOLD},
+        {"value": 800, "name": WEIGHT_EXTRA_BOLD},
+        {"value": 900, "name": WEIGHT_BLACK},
+        {"value": 950, "name": WEIGHT_EXTRA_BLACK},
+    ]
+    _WEIGHTS_BY_VALUE = {weight["value"]: weight for weight in _WEIGHTS}
 
     # Widths:
     # https://docs.microsoft.com/en-us/typography/opentype/otspec170/os2#uswidthclass
@@ -187,17 +192,18 @@ class Font(object):
     WIDTH_EXPANDED = "Expanded"
     WIDTH_EXTRA_EXPANDED = "Extra-expanded"
     WIDTH_ULTRA_EXPANDED = "Ultra-expanded"
-    _WIDTHS_BY_VALUE = {
-        1: WIDTH_ULTRA_CONDENSED,
-        2: WIDTH_EXTRA_CONDENSED,
-        3: WIDTH_CONDENSED,
-        4: WIDTH_SEMI_CONDENSED,
-        5: WIDTH_MEDIUM,
-        6: WIDTH_SEMI_EXPANDED,
-        7: WIDTH_EXPANDED,
-        8: WIDTH_EXTRA_EXPANDED,
-        9: WIDTH_ULTRA_EXPANDED,
-    }
+    _WIDTHS = [
+        {"value": 1, "perc": 50.0, "name": WIDTH_ULTRA_CONDENSED},
+        {"value": 2, "perc": 62.5, "name": WIDTH_EXTRA_CONDENSED},
+        {"value": 3, "perc": 75.0, "name": WIDTH_CONDENSED},
+        {"value": 4, "perc": 87.5, "name": WIDTH_SEMI_CONDENSED},
+        {"value": 5, "perc": 100.0, "name": WIDTH_MEDIUM},
+        {"value": 6, "perc": 112.5, "name": WIDTH_SEMI_EXPANDED},
+        {"value": 7, "perc": 125.0, "name": WIDTH_EXPANDED},
+        {"value": 8, "perc": 150.0, "name": WIDTH_EXTRA_EXPANDED},
+        {"value": 9, "perc": 200.0, "name": WIDTH_ULTRA_CONDENSED},
+    ]
+    _WIDTHS_BY_VALUE = {width["value"]: width for width in _WIDTHS}
 
     def __init__(self, filepath):
         """
@@ -218,7 +224,7 @@ class Font(object):
         else:
             filepath_type = type(filepath).__name__
             raise ValueError(
-                f"Invalid filepath type: expected str, found {filepath_type}"
+                f"Invalid filepath type: expected str, found '{filepath_type}'."
             )
 
     def _init_with_filepath(self, filepath):
@@ -226,7 +232,7 @@ class Font(object):
             self._ttfont = TTFont(filepath)
             self._filepath = filepath
         except TTLibError:
-            raise ValueError(f"Invalid font at filepath: {filepath}")
+            raise ValueError(f"Invalid font at filepath: '{filepath}'.")
 
     def __enter__(self):
         return self
@@ -314,6 +320,23 @@ class Font(object):
             return features_tags
         return []
 
+    def get_fingerprint(self, text=""):
+        """
+        Gets the font fingerprint: an hash calculated from an image representation of the font.
+        Changing the text option affects the returned fingerprint.
+
+        :param text: The text used for generating the fingerprint, default value: "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
+        :type text: str
+
+        :returns: The fingerprint hash.
+        :rtype: imagehash.ImageHash
+        """
+        text = text or "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
+        img = self.get_image(text=text, size=72)
+        # img.show()
+        hash = imagehash.dhash(img, hash_size=32)
+        return hash
+
     def get_format(self, ignore_flavor=False):
         """
         Gets the font format: otf, ttf, woff, woff2.
@@ -340,6 +363,53 @@ class Font(object):
             format = self.FORMAT_WOFF2
         return format
 
+    def get_image(
+        self, text, size, color=(0, 0, 0, 255), background_color=(255, 255, 255, 255)
+    ):
+        """
+        Gets an image representation of the font rendering
+        some text using the given options.
+
+        :param text: The text rendered in the image
+        :type text: str
+        :param size: The font size
+        :type size: int
+        :param color: The text color
+        :type color: tuple
+        :param background_color: The background color
+        :type background_color: tuple
+        """
+        with tempfile.TemporaryDirectory() as dest:
+            filepath = self.save(dest)
+            img = Image.new("RGBA", (2, 2), background_color)
+            draw = ImageDraw.Draw(img)
+            img_font = ImageFont.truetype(filepath, size)
+            img_size = draw.textsize(text, img_font)
+            img = img.resize(img_size)
+            draw = ImageDraw.Draw(img)
+            draw.text((0, 0), text, font=img_font, fill=color)
+            return img
+
+    def get_italic_angle(self):
+        """
+        Gets the font italic angle.
+
+        :returns: The angle value including backslant, italic and roman flags.
+        :rtype: dict or None
+        """
+        font = self.get_ttfont()
+        post = font.get("post")
+        if not post:
+            return None
+        italic_angle_value = post.italicAngle
+        italic_angle = {
+            "backslant": italic_angle_value > 0,
+            "italic": italic_angle_value < 0,
+            "roman": italic_angle_value == 0,
+            "value": italic_angle_value,
+        }
+        return italic_angle
+
     @classmethod
     def _get_name_id(cls, key):
         if isinstance(key, int):
@@ -348,7 +418,7 @@ class Font(object):
             return cls._NAMES_BY_KEY[key]["id"]
         else:
             raise TypeError(
-                f"Invalid key type, expected int or str, found {type(key).__name__}."
+                f"Invalid key type, expected int or str, found '{type(key).__name__}'."
             )
 
     def get_name(self, key):
@@ -663,12 +733,16 @@ class Font(object):
         os2 = font.get("OS/2")
         if not os2:
             return None
-        weight = os2.usWeightClass
-        weight = min(max(0, weight), 1000)
-        weights = sorted(self._WEIGHTS_BY_VALUE.keys())
-        closest_weight = min(weights, key=lambda weight_item: abs(weight_item - weight))
-        weight_name = self._WEIGHTS_BY_VALUE.get(closest_weight)
-        return {"name": weight_name, "value": weight}
+        weight_value = os2.usWeightClass
+        weight_value = min(max(1, weight_value), 1000)
+        weight_option_values = sorted(self._WEIGHTS_BY_VALUE.keys())
+        closest_weight_option_value = min(
+            weight_option_values,
+            key=lambda weight_option_value: abs(weight_option_value - weight_value),
+        )
+        weight = self._WEIGHTS_BY_VALUE.get(closest_weight_option_value, {}).copy()
+        weight["value"] = weight_value
+        return weight
 
     def get_width(self):
         """
@@ -681,10 +755,11 @@ class Font(object):
         os2 = font.get("OS/2")
         if not os2:
             return None
-        width = os2.usWidthClass
-        width = min(max(0, width), 9)
-        width_name = self._WIDTHS_BY_VALUE.get(width)
-        return {"name": width_name, "value": width}
+        width_value = os2.usWidthClass
+        width_value = min(max(1, width_value), 9)
+        width = self._WIDTHS_BY_VALUE.get(width_value, {}).copy()
+        width["value"] = width_value
+        return width
 
     def is_static(self):
         """
@@ -705,6 +780,36 @@ class Font(object):
         font = self.get_ttfont()
         return "fvar" in font
 
+    def match(self, other, tolerance=5):
+        """
+        Check if the font has the same "identity" of another font
+        by checking if the fingerprints difference is <= tolerance.
+
+        :param other: The other font, can be either a filepath or a Font instance.
+        :type other: str or Font
+        :param tolerance: The diff tolerance, default 3.
+        :type tolerance: int
+
+        :returns: True if match other font, False otherwise.
+        :rtype: bool
+        """
+        other_font = None
+        if isinstance(other, str):
+            other_font = Font(other)
+        elif isinstance(other, Font):
+            other_font = other
+        else:
+            other_type = type(other).__name__
+            raise ValueError(
+                f"Invalid other filepath/font: expected str or Font instance, found '{other_type}'."
+            )
+        if self.is_variable() != other_font.is_variable():
+            return False
+        hash = self.get_fingerprint()
+        other_hash = other_font.get_fingerprint()
+        hash_diff = hash - other_hash
+        return hash_diff <= tolerance
+
     def rename(self, family_name="", style_name=""):
         """
         Renames the font names records (1, 2, 4, 6, 16, 17) according to
@@ -976,7 +1081,8 @@ class Font(object):
         If an axis value is not specified, the axis will be left untouched.
         If an axis min and max values are equal, the axis will be pinned.
 
-        :param coordinates: The coordinates dictionary, each item value must be tuple/list/dict (with min and max keys) for slicing or float/int for pinning, eg.
+        :param coordinates: The coordinates dictionary, each item value must be
+            tuple/list/dict (with min and max keys) for slicing or float/int for pinning, eg.
             {'wdth':100, 'wght':(100,600), 'ital':(30,70)} or
             {'wdth':100, 'wght':[100,600], 'ital':[30,70]} or
             {'wdth':100, 'wght':{'min':100,'max':600}, 'ital':{'min':30,'max':70}}
